@@ -11,13 +11,15 @@ from deposit.utils.fnc_files import (
 	clear_temp_dir, get_named_path, get_free_subfolder, get_unique_path, 
 	prune_local_files,
 )
+from deposit.utils.fnc_serialize import (value_to_str)
+from deposit.utils.fnc_get_linked_objects import (get_linked_objects)
 
 from collections import defaultdict
 from natsort import natsorted
-from itertools import product
 import shutil
 import os
 
+
 class Store(object):
 	
 	def __init__(self, keep_temp = False, *args, **kwargs):
@@ -261,15 +263,28 @@ class Store(object):
 		self._added = set()
 		self._deleted = set()
 	
-	def get_updated_url(self, resource):
+	def get_updated_url(self, resource, store = None):
 		
+		if store is None:
+			store = self
 		if not resource.is_stored:
 			return resource.url
-		url_new = get_updated_local_url(resource.url, self.get_folder())
+		url_new = get_updated_local_url(resource.url, store.get_folder())
 		if url_new != resource.url:
+			if resource.url in self._resources:
+				del self._resources[resource.url]
 			resource.value = (url_new, resource.filename, resource.is_stored, resource.is_image)
+			self._resources[url_new] = resource
 		return url_new
 	
+	def update_resource_urls(self):
+		
+		for obj in self.get_objects():
+			for descr in obj.get_descriptors():
+				value = obj.get_descriptor(descr)
+				if isinstance(value, DResource):
+					self.get_updated_url(value)
+	
 	def set_local_folder(self, path, progress = None):
 		
 		if path is None:
@@ -445,103 +460,223 @@ class Store(object):
 			return
 		self._queries[title] = querystr
 		self.callback_queries_changed()
-	
-	def find_object_with_descriptors(self, classes, data, locations = {}):
-		# classes = [DClass, None, ...]
-		# data = {descriptor_name: value, ...}
-		# locations = {descriptor_name: location, ...}
-		#
-		# if found object has missing location, it is updated if a location is
-		# supplied via locations
+
+	def _is_equal(self, value1, value2):
 		
-		def _get_class_members(cls):
+		if isinstance(value1, DResource) and isinstance(value2, DResource) and value1.is_stored and value2.is_stored:
+			return value1.filename == value2.filename
+		return value_to_str(value1) == value_to_str(value2)
+	
+	def _check_related(self, obj, related_data, match_empty=True, checked_objects=None, matches=None):
+		
+		if not related_data:
+			return 0
+		
+		if matches is None:
+			matches = set()
+		
+		if checked_objects is None:
+			checked_objects = set()
+	
+		if obj in checked_objects:
+			return -1
+		
+		for (class_name_1, label, class_name_2, descriptor_name), expected_value in related_data.items():
+			key = (class_name_1, label, class_name_2, descriptor_name)
+			related_objects = [
+				rel_obj for rel_obj, rel_label in obj.get_relations() if rel_label == label
+			]
+			found_one = False
+			for rel_obj in related_objects:
+				if class_name_2 in rel_obj.get_class_names():
+					if descriptor_name in rel_obj.get_descriptor_names():
+						found_one = True
+						value = rel_obj.get_descriptor(descriptor_name)
+						if not self._is_equal(value, expected_value):
+							continue
+						matches.add(key)
+						# Check for nested related data specific to the found object
+						nested_related_data = {
+							k: v for k, v in related_data.items() if k[0] == class_name_2
+						}
+						if nested_related_data:
+							if self._check_related(rel_obj, nested_related_data, match_empty, checked_objects.copy(), matches) == -1:
+								return -1
+			if (match_empty or found_one) and (key not in matches):
+				return -1
+		checked_objects.add(obj)
+		return len(matches)
+	
+	def _check_descriptors_and_locations(self, obj, data, locations, match_empty=True):
+		
+		matches = set()	
+		# Check descriptor values
+		for descriptor_name in data:
+			value = obj.get_descriptor(descriptor_name)
+			if self._is_equal(value, data[descriptor_name]):
+				matches.add(descriptor_name)
 			
+			if (not self._is_equal(value, data[descriptor_name])) and (match_empty or ((value is not None) and (data[descriptor_name] is not None))):
+				return -1
+			
+			# Check descriptor location
+			if not self._check_location(obj, descriptor_name, locations, match_empty):
+				return -1
+		
+		return len(matches)
+	
+	def _check_location(self, obj, descriptor_name, locations, match_empty=True):
+		# Helper function for find_object_with_descriptors
+		if descriptor_name not in locations:
+			return True
+		location = obj.get_location(descriptor_name)
+		if (location is None) and (not match_empty):
+			return True
+		if location == locations[descriptor_name]:
+			return True
+		return False
+	
+	def find_object_with_descriptors(self, classes, data, locations={}, related_data={}, match_empty=True):
+		"""
+		Finds an object that meets specified criteria within given classes, descriptors,
+		locations, and related objects.
+
+		Parameters:
+		-----------
+		classes : [DClass, None, ...]
+			A list of class objects to search within. Use `None` to include all objects
+			with any class.
+		data : {descriptor_name: value, ...}
+			A dictionary mapping descriptor names to their required values for the search.
+		locations : {descriptor_name: location, ...}, optional
+			A dictionary specifying the locations for each descriptor.
+		related_data : {(class_name_1, label, class_name_2, descriptor_name): value, ...}, optional
+			A dictionary defining the relationships to other objects that must be checked.
+			Specifies which class, descriptor, and label are required in the related objects.
+		match_empty : bool, optional
+			If True, check missing descriptors, locations and relations. If False, add missing locations and descriptors
+
+		Returns:
+		--------
+		object or None
+			The first object that meets all specified criteria. Returns `None` if no such
+			object is found.
+		"""
+
+		def _get_class_members(cls):
 			if cls is None:
 				return set([obj for obj in self.get_objects() if obj.get_classes()])
-			return cls.get_members(direct_only = True)
+			return cls.get_members(direct_only=True)
 		
-		def _check_location(obj, descriptor_name, locations):
+		def _check_object(obj, data, locations, related_data, match_empty):
+			matches_descr = self._check_descriptors_and_locations(obj, data, locations, match_empty)
+			if matches_descr < 0:
+				return 0
+		
+			# Check related objects
+			matches_rel = self._check_related(obj, related_data, match_empty)
+			if matches_rel < 0:
+				return 0
 			
-			if descriptor_name not in locations:
-				return True
-			location = obj.get_location(descriptor_name)
-			if location is None:
-				obj.set_location(descriptor_name, locations[descriptor_name])
-				return True
-			if location == locations[descriptor_name]:
-				return True
-			return False
+			return matches_descr + matches_rel
 		
+		found = []
+		if not classes:
+			classes = [None]
 		for obj in set.union(*[_get_class_members(cls) for cls in classes]):
-			found = True
-			for descriptor_name in data:
-				if (obj.get_descriptor(descriptor_name) != data[descriptor_name]) and \
-				(_check_location(obj, descriptor_name, locations)):
-					found = False
-					break
-			if found:
-				return obj
+			matches = _check_object(obj, data, locations, related_data, match_empty)
+			if matches > 0:
+				found.append((obj, matches))
+		if not found:
+			return None
 		
-		return None
+		# Select match with the most matches
+		obj, _ = sorted(found, key = lambda item: item[1], reverse=True)[0]
+		
+		return obj
 	
-	def add_external_descriptor(self, obj, name, value):
+	def add_external_descriptor(self, obj, name, value, store = None):
 		# localize descriptor if it is a DResource
 		
+		if store is None:
+			store = self
 		if isinstance(value, DResource):
 			is_stored = value.is_stored
+			url = self.get_updated_url(value, store = store)
 			if is_stored == True:
 				is_stored = None
 			obj.set_resource_descriptor(name, value.url, value.filename, is_stored, value.is_image)
 		else:
 			obj.set_descriptor(name, value)
 	
-	def add_object_with_descriptors(self, cls, data, locations = {}, obj = None):
+	def set_object_descriptors(self, obj, data, locations = {}, match_empty=True, store = None):
+		if store is None:
+			store = self
+		for descriptor_name in data:
+			value = data[descriptor_name]
+			if (not match_empty) and (value is None) and obj.has_descriptor(descriptor_name):
+				continue
+			self.add_external_descriptor(obj, descriptor_name, value, store = store)
+			if descriptor_name in locations:
+				obj.set_location(descriptor_name, locations[descriptor_name])
+	
+	def add_object_with_descriptors(self, cls, data, locations = {}, obj = None, store = None):
 		# cls = DClass or None
 		# data = {descriptor_name: value, ...}
 		# locations = {descriptor_name: location, ...}
 		
+		if store is None:
+			store = self
 		if obj is None:
 			obj = self.add_object()
 		if cls is not None:
 			cls.add_member(obj.id)
-		for descriptor_name in data:
-			value = data[descriptor_name]
-			self.add_external_descriptor(obj, descriptor_name, value)
-			if descriptor_name in locations:
-				obj.set_location(descriptor_name, locations[descriptor_name])
+		self.set_object_descriptors(obj, data, locations=locations, match_empty=True, store=store)
 		
 		return obj
 	
-	def add_data_row(self, 
-		data: dict, 
-		relations: set = set(), 
-		unique: set = set(), 
-		existing = {}, 
-		return_added = False,
-	):
-		# add multiple objects with classes at once & automatically add relations 
-		#	based on class relations or as specified in the relations attribute
-		# data = {(Class name, Descriptor name): value, ...}
-		# relations = {(Class name 1, label, Class name 2), ...}
-		# unique = {Class name, ...}; always add a new object to classes 
-		#	specified here, otherwise re-use objects with identical descriptors
-		# existing = {Class name: Object, ...}
-		#	use existing object for specified classes (i.e. just update descriptors)
-		#
-		# returns n_added or (n_added, added) if return_added == True
-		#	added = {Class name: Object, ...}
+	def add_data_row(self, data, relations=set(), unique=set(), existing={}, return_added=False, match_empty=False):
+		"""
+		Adds objects to the store with specified classes, descriptors, and relations.
+		
+		Parameters:
+		-----------
+		data : {(Class name, Descriptor name): value, ...}
+			A dictionary where the keys are tuples of (Class name, Descriptor name) and the values 
+			are the corresponding descriptor values to be set for the objects.
+		relations : set([(Class name 1, label, Class name 2), ...]), optional
+			A set of tuples defining the relationships between classes. Each tuple specifies a 
+			relationship in the form (Class name 1, Label, Class name 2).
+		unique : set([Class name, ...]), optional
+			A set of class names where new objects should always be added, rather than reusing 
+			existing objects with identical descriptors. This ensures unique objects for these classes.
+		existing : {Class name: Object, ...}, optional
+			A dictionary mapping class names to existing objects that should be updated rather than 
+			creating new ones. 
+		match_empty : bool, optional
+			If True, check missing locations and descriptors. If False, add missing locations and descriptors
+		
+		return_added : bool, optional
+			If `True`, the function returns a tuple containing the number of added objects and a 
+			dictionary mapping class names to the added objects. Otherwise, it only returns the 
+			number of added objects.
+		
+		Returns:
+		--------
+		int or tuple
+			Returns the number of objects added. If `return_added` is `True`, it returns a tuple 
+			(n_added, added) where `n_added` is the number of objects added, and `added` is a 
+			dictionary mapping class names to the added objects.
+		"""
 		
 		collect = defaultdict(dict)
 		for key in data:
-			if data[key] is None:
-				continue
 			class_name, descriptor_name = key
 			if (class_name is None) or (descriptor_name is None):
 				continue
 			cls = self.add_class(class_name)
-			descr = self.add_class(descriptor_name)
 			collect[cls][descriptor_name] = data[key]
-		data = collect  # {cls: {descriptor_name: value, ...}, ...}
+		data = collect	# {cls: {descriptor_name: value, ...}, ...}
 		
 		collect = defaultdict(set)
 		for class_name1, label, class_name2 in relations:
@@ -561,21 +696,82 @@ class Store(object):
 			for cls2, label in cls1.get_relations():
 				if cls2 in data:
 					collect[cls1].add((label, cls2))
-		relations_all = collect  # {cls1: [(label, cls2), ...], ...}
+		relations_all = collect	 # {cls1: [(label, cls2), ...], ...}
+		
+		relations_one_way = set()
+		for cls1 in relations_all:
+			for label, cls2 in relations_all[cls1]:
+				if label.startswith("~"):
+					continue
+				if cls1 == cls2:
+					continue
+				relations_one_way.add((cls1, label, cls2))
+		relation_tiers = []
+		done = set()
+		classes_done = set()
+		while True:
+			todo = relations_one_way.difference(done)
+			if not todo:
+				break
+			tier_relations = set()
+			tier_classes = set()
+			on_right = set([row[2] for row in todo])
+			for row in todo:
+				if row[0] not in on_right:
+					tier_classes.add(row[0])
+					for row2 in relations_one_way:
+						if row2[2] == row[0]:
+							tier_relations.add(row2)
+					done.add(row)
+			if tier_classes:
+				relation_tiers.append((tier_relations, tier_classes))
+				classes_done.update(tier_classes)
+			else:
+				break
+		
+		for cls in data:
+			if cls in classes_done:
+				continue
+			tier_relations = set()
+			tier_classes = set([cls])
+			classes_done.add(cls)
+			for row in relations_one_way:
+				if row[2] == cls:
+					tier_relations.add(row)
+			relation_tiers.append((tier_relations, tier_classes))
 		
 		n_added = 0
-		added = {}  # {cls: obj, ...}
-		for cls in data:
-			added[cls] = None
-			if cls.name in existing:
-				added[cls] = self.add_object_with_descriptors(
-					cls, data[cls], obj = existing[cls.name]
-				)
-			elif cls.name not in unique:
-				added[cls] = self.find_object_with_descriptors([cls], data[cls])
-			if added[cls] is None:
-				added[cls] = self.add_object_with_descriptors(cls, data[cls])
-				n_added += 1
+		added = {}	# {cls: obj, ...}
+		for i, (tier_relations, tier_classes) in enumerate(relation_tiers):
+			for cls in tier_classes:
+				added[cls] = None
+				if cls.name in existing:
+					added[cls] = self.add_object_with_descriptors(
+						cls, data[cls], obj=existing[cls.name]
+					)
+				elif cls.name not in unique:
+					related_data = {}  # {(class_name_1, label, class_name_2, descriptor_name): value, ...}
+					cls_done = set()
+					for cls1, label, cls2 in tier_relations:
+						if cls2 != cls:
+							continue
+						if cls1 in data:
+							for descriptor_name in data[cls1]:
+								related_data[(cls2.name, "~" + label, cls1.name, descriptor_name)] = data[cls1][descriptor_name]
+								cls_done.add(cls1)
+					for tier_relations_prev, tier_classes_prev in relation_tiers[:i]:
+						for cls1, label, cls2 in tier_relations_prev:
+							if cls2 not in cls_done:
+								continue
+							if cls1 in data:
+								for descriptor_name in data[cls1]:
+									related_data[(cls2.name, "~" + label, cls1.name, descriptor_name)] = data[cls1][descriptor_name]
+					added[cls] = self.find_object_with_descriptors([cls], data[cls], related_data=related_data, match_empty=match_empty)
+				if added[cls] is None:
+					added[cls] = self.add_object_with_descriptors(cls, data[cls])
+					n_added += 1
+				else:
+					self.set_object_descriptors(added[cls], data[cls], match_empty=match_empty)
 		
 		for cls1 in relations_all:
 			if cls1 not in added:
@@ -589,63 +785,177 @@ class Store(object):
 			return n_added, dict([(cls.name, added[cls]) for cls in added])
 		return n_added
 	
-	def import_store(self, store, unique: set = set(), progress = None) -> None:
-		# unique = {Class name, ...}; always add a new object to classes 
-		#	specified here, otherwise re-use objects with identical descriptors
-		# progress = DProgress
+	def import_store(self, store, unique: set=set(), match_empty: bool=False, progress=None) -> None:
+		"""
+		Imports objects from another store instance recursively and tiered by dependency.
 		
-		for cls in store.get_classes(ordered = True):
+		Parameters:
+		-----------
+		store : Store
+			The store instance to import objects from.
+		unique : set, optional
+			A set of class names for which new objects should always be added rather than reusing existing ones.
+		match_empty : bool, optional
+			If True, check missing locations and descriptors. If False, add missing locations and descriptors
+		progress : DProgress, optional
+			Progress indicator.
+		"""
+		def _should_cancel():
+			if progress and progress.cancel_pressed():
+				return True
+			return False
+		
+		def _get_related_data(obj, tier_relations, store):
+			# get data from ancestors of obj where (cls2, label, cls1) is in tier_relations
+			related_data = {}  # {(cls1, ~label, cls2, descriptor_name): value, ...}
+			queue = set([obj.id])
+			done = set()
+			while queue:
+				orig_id = queue.pop()
+				orig = store.get_object(orig_id)
+				classes1 = set(orig.get_class_names())
+				for orig2, label in orig.get_relations():
+					if not label.startswith('~'):
+						continue
+					classes2 = set(orig2.get_class_names())
+					if classes2.intersection(classes1):
+						continue
+					if orig2.id in done:
+						continue
+					cls_combos = set()
+					for cls1 in classes1:
+						for cls2 in classes2:
+							if (cls2, label[1:], cls1) in tier_relations:
+								cls_combos.add((cls1, cls2))
+					if not cls_combos:
+						continue
+					for descr in orig2.get_descriptor_names():
+						for cls1, cls2 in cls_combos:
+							related_data[(cls1, label, cls2, descr)] = orig2.get_descriptor(descr)
+					queue.add(orig2.id)
+					done.add(orig2.id)
+			return related_data
+		
+		def _get_object_data(obj):
+			classes = [self.get_class(cls.name) for cls in obj.get_classes()]
+			data = {descr.name: obj.get_descriptor(descr) for descr in obj.get_descriptors()}
+			locations = {descr.name: obj.get_location(descr) for descr in obj.get_descriptors() if obj.get_location(descr) is not None}
+			return classes, data, locations
+		
+		def _add_object(classes):
+			obj = self.add_object()
+			for cls in classes:
+				self.G.add_class_child(cls.name, obj.id)
+			return obj
+		
+		# Step 1: Import classes and descriptors
+		for cls in store.get_classes(ordered=True):
 			self.add_class(cls.name)
 		for cls in store.get_classes():
 			cls_ = self.get_class(cls.name)
-			for descr in cls.get_descriptors(ordered = True):
+			for descr in cls.get_descriptors(ordered=True):
 				cls_.set_descriptor(descr.name)
 			for subcls in cls.get_subclasses():
 				cls_.add_subclass(subcls)
 			for cls2, label in cls.get_relations():
 				cls_.add_relation(cls2.name, label)
 		
-		objects = store.get_objects()
-		cmax = 2*len(objects)
-		cnt = 1
-		added = {}  # {orig_obj: new_obj, ...}
-		for obj in objects:
-			if progress is not None:
-				if progress.cancel_pressed():
-					return
-				progress.update_state(text = "Importing", value = cnt, maximum = cmax)
-				cnt += 1
-			data = {}
-			locations = {}
-			for descr in obj.get_descriptors():
-				data[descr.name] = obj.get_descriptor(descr)
-				location = obj.get_location(descr)
-				if location is not None:
-					locations[descr.name] = location
-			classes = [self.get_class(cls.name) for cls in obj.get_classes()]
-			if not classes:
-				classes = [None]
-			obj_ = None
-			if True not in [(cls.name in unique) for cls in classes if cls is not None]:
-				obj_ = self.find_object_with_descriptors(classes, data, locations)
-			if obj_ is None:
-				obj_ = self.add_object_with_descriptors(classes[0], data, locations)
-			for cls in classes:
-				if cls is None:
+		# Step 2: Prepare for tiered import
+		relations = set()
+		for cls1 in store.get_classes():
+			for cls2, label in cls1.get_object_relations(direct_only = True):
+				if label.startswith("~"):
 					continue
-				cls.add_member(obj_.id)
-			added[obj] = obj_
+				if cls1 == cls2:
+					continue
+				relations.add((cls1.name, label, cls2.name))
+		relation_tiers = []
+		done = set()
+		classes_done = set()
+		while True:
+			todo = relations.difference(done)
+			if not todo:
+				break
+			tier_relations = set()
+			tier_classes = set()
+			on_right = set([row[2] for row in todo])
+			for row in todo:
+				if row[0] not in on_right:
+					tier_classes.add(row[0])
+					for row2 in relations:
+						if row2[2] == row[0]:
+							tier_relations.add(row2)
+					done.add(row)
+			if tier_classes:
+				relation_tiers.append((tier_relations, tier_classes))
+				classes_done.update(tier_classes)
+			else:
+				break
 		
-		for obj in added:
-			if progress is not None:
-				if progress.cancel_pressed():
-					return
-				progress.update_state(text = "Importing", value = cnt, maximum = cmax)
-				cnt += 1
-			obj_ = added[obj]
-			for obj2, label in obj.get_relations():
-				weight = obj.get_relation_weight(obj2, label)
-				obj_.add_relation(added[obj2], label, weight = weight)
+		for cls in store.get_classes():
+			if cls in classes_done:
+				continue
+			tier_relations = set()
+			tier_classes = set([cls.name])
+			classes_done.add(cls.name)
+			for row in relations:
+				if row[2] == cls:
+					tier_relations.add(row)
+			relation_tiers.append((tier_relations, tier_classes))
+		
+		# Find or add objects based on their classes, descriptors, locations and related data
+		tier_relations = set()
+		obj_lookup = {}
+		objects_done = set()
+		cnt = 0
+		cmax = len(relation_tiers)
+		for tier_relations_, tier_classes in relation_tiers:
+			tier_relations.update(tier_relations_)
+			# tier_relations = {(cls1, label, cls2), ...}
+			# tier_classes = {cls, ...}
+			if _should_cancel():
+				return
+			if progress:
+				progress.update_state(text="Importing", value=cnt, maximum=cmax)
+			cnt += 1
+			for cls in tier_classes:
+				for obj_src in store.get_class(cls).get_members(direct_only=True):
+					if obj_src.id in obj_lookup:
+						continue
+					related_data = _get_related_data(obj_src, tier_relations, store)  # {(cls1, ~label, cls2, descriptor_name): value, ...}
+					classes, data, locations = _get_object_data(obj_src)
+					obj = self.find_object_with_descriptors(classes, data, locations, related_data, match_empty=match_empty)
+					if obj is None:
+						obj = _add_object(classes)
+					if obj.id not in objects_done:
+						self.set_object_descriptors(obj, data, locations=locations, match_empty=match_empty, store=store)
+						objects_done.add(obj.id)
+					obj_lookup[obj_src.id] = obj
+		
+		# Add any remaining objects
+		for obj_src in store.get_objects():
+			if obj_src.id not in obj_lookup:
+				classes, data, locations = _get_object_data(obj_src)
+				obj = _add_object(classes)
+				obj_lookup[obj_src.id] = obj
+				self.set_object_descriptors(obj, data, locations=locations, match_empty=match_empty, store=store)
+				objects_done.add(obj.id)
+		for obj_src_id in obj_lookup:
+			obj = obj_lookup[obj_src_id]
+			if obj.id not in objects_done:
+				classes, data, locations = _get_object_data(store.get_object(obj_src_id))
+				self.set_object_descriptors(obj, data, locations=locations, match_empty=match_empty, store=store)
+		
+		# Add relations between objects
+		for obj_src1 in store.get_objects():
+			for obj_src2, label in obj_src1.get_relations():
+				if label.startswith('~'):
+					continue
+				weight = obj_src1.get_relation_weight(obj_src2.id, label)
+				obj_lookup[obj_src1.id].add_relation(obj_lookup[obj_src2.id], label, weight)
+		
+		if progress:
+			progress.update_state(text="Importing Complete", value=cmax, maximum=cmax)
 	
 	
 	# ---- Read
@@ -778,255 +1088,11 @@ class Store(object):
 			return label[1:]
 		return "~" + label
 	
-	def get_class_connections(self, classes, relations, progress = None):
+	def get_linked_objects(self, classes, relations, progress=None):
 		# classes = [class_name, ...]
 		# relations = [(class1, label, class2), ...]
 		
-		def _get_class_members(class_name, leave_names = False):
-			
-			if class_name == "*":
-				return set([obj.id for obj in self.get_objects()])
-			if class_name == "!*":
-				return set([obj.id for obj in self.get_objects() if not obj.has_class()])
-			if isinstance(class_name, tuple):
-				return set(list(class_name))
-			cls = self.get_class(class_name)
-			if cls is None:
-				return set()
-			if leave_names:
-				return class_name
-			return set([obj.id for obj in cls.get_members()])
-		
-		def _get_class_names(cls):
-			
-			if cls == "*":
-				return set([cls_.name for cls_ in self.get_classes()])
-			if cls == "!*":
-				return set(["!*"])
-			if isinstance(cls, tuple):
-				clss = set.union(*[set([cls_.name for cls_ in \
-					self.get_object(obj_id).get_classes()]) for \
-						obj_id in list(cls)])
-				if not clss:
-					clss = set(["!*"])
-				return clss
-			return set([cls])
-		
-		def _get_within_class_rel(cls1, cls2):
-			
-			clss1 = _get_class_names(cls1)
-			clss2 = _get_class_names(cls2)
-			return clss1.intersection(clss2)
-		
-		def _get_paths(
-			obj_id0, cls_lookup, rels, within_cls_rels, 
-			asterisk_rels, classes, connecting, mandatory_classes,
-		):
-			
-			'''
-			Get all paths from obj_id0 to objects from specified classes.
-			1. If classes or relations contains object specifications OBJ(1, ...)
-			require each path to contain one of those objects
-			2. If a cls1.*.cls2 relation is specified, the path must contain an 
-			edge between cls1 and cls2.
-			3. The paths cannot have within-class edges unless explicitly
-			specified in relations.
-			4. If a path contains an edge between classes as specified in 
-			relations, it must have a label as specified in relations.
-			'''
-			queue = [[[obj_id0], set(), cls_lookup[obj_id0].copy(), False, False]]
-			# queue = [[path, found_asterisk, found_classes, rule1, rule3], ...]
-			done = set()
-			skip = set()
-			while queue:
-				path, found_asterisk, found_classes, rule1, rule3 = queue.pop()
-				found_next = False
-				if not mandatory_classes:
-					rule1 = True
-				if (len(found_classes) < len(classes)) or within_cls_rels:
-					
-					src = path[-1]
-					clss_src = cls_lookup.get(src, None)
-					
-					# rule 1 - first stage, source
-					if (not rule1) and (clss_src is not None) and \
-						clss_src.intersection(mandatory_classes):
-							rule1 = True
-					
-					for tgt, label in self.G.iter_object_relations(src):
-						
-						# already processed
-						if tgt in skip:
-							continue
-						skip.add(tgt)
-						
-						# circular path
-						if tgt in path:
-							continue
-						
-						clss_tgt = cls_lookup.get(tgt, None)
-						
-						# only check objects with specified classes or from the connecting set
-						if (clss_tgt is None) and (tgt not in connecting):
-							continue
-						
-						# collect all combinations of source & target classes
-						clss = []
-						if (clss_src is not None) and (clss_tgt is not None):
-							if rels or asterisk_rels:
-								clss = list(product(clss_src, clss_tgt))
-						
-						# rule 2 - first stage
-						if clss and asterisk_rels:
-							for cls_src, cls_tgt in clss:
-								if (cls_src, cls_tgt) in asterisk_rels:
-									found_asterisk.add((cls_src, cls_tgt))
-						
-						# rule 3 - first stage
-						found_classes_ = found_classes.copy()
-						if clss_tgt is not None:
-							common = found_classes_.intersection(clss_tgt)
-							if common:
-								if (not rule3) and (clss_src is not None):
-									for cls in clss_src:
-										if (cls in within_cls_rels) and (within_cls_rels[cls].intersection([label, "*"])):
-											rule3 = True
-											break
-								continue
-						
-						# rule 1 - first stage, target
-						if clss_tgt is not None:
-							if (not rule1) and clss_tgt.intersection(mandatory_classes):
-								rule1 = True
-						
-						# rule 4
-						if clss:
-							found_rel = False
-							found_label = False
-							if rels:
-								for cls_src, cls_tgt in clss:
-									if (cls_src, cls_tgt) in rels:
-										found_rel = True
-										if rels[(cls_src, cls_tgt)].intersection(
-											[label, "*"]
-										) or rels[(cls_tgt, cls_src)].intersection(
-											[self.reverse_relation(label), "*"]
-										):
-											found_label = True
-											break
-							if found_rel and not found_label:
-								continue
-						
-						if clss_tgt is not None:
-							found_classes_.update(clss_tgt)
-						
-						queue.append([path + [tgt], found_asterisk, found_classes_, rule1, rule3])
-						found_next = True
-				
-				if found_next:
-					continue
-				
-				# rule 2 - second stage
-				rule2 = True
-				for cls1, cls2 in asterisk_rels:
-					if not (((cls1, cls2) in found_asterisk) or \
-						((cls2, cls1) in found_asterisk)):
-							rule2 = False
-							break
-				if not rule2:
-					continue
-				
-				# rule 3 - second stage
-				if within_cls_rels and not rule3:
-					continue
-				
-				# rule 1 - second stage
-				if mandatory_classes and not rule1:
-					continue
-				
-				# only add objects from specified classes
-				done.add(tuple(sorted(
-					[obj_id for obj_id in path if obj_id in cls_lookup]
-				)))
-				
-			return done
-		
-		rels = defaultdict(set)  # {(class1, class2): set(label, ...), ...}
-		within_cls_rels = defaultdict(set)  # {class: set(label, ...), ...}
-		asterisk_rels = set()  # set((class1, class2), ...)
-		mandatory_classes = set() # set(class, ...)
-		for cls1, lbl, cls2 in relations:
-			if isinstance(cls1, tuple):
-				mandatory_classes.add(cls1)
-			if isinstance(cls2, tuple):
-				mandatory_classes.add(cls2)
-			if lbl == "*":
-				asterisk_rels.add((cls1, cls2))
-				asterisk_rels.add((cls2, cls1))
-			clss = _get_within_class_rel(cls1, cls2)
-			if clss:
-				within_cls_rels[cls1].add(lbl)
-			else:
-				rels[(cls1, cls2)].add(lbl)
-				rels[(cls2, cls1)].add(self.reverse_relation(lbl))
-		
-		cls0 = classes[0]
-		objects0 = _get_class_members(cls0)
-		if not objects0:
-			# no objects found in the first specified class
-			return set()
-		
-		cls_lookup = defaultdict(set)
-		collect = []
-		for cls in classes:
-			if isinstance(cls, tuple):
-				mandatory_classes.add(cls)
-			objs = objects0 if (cls == cls0) else _get_class_members(cls)
-			if objs:
-				collect.append(cls)
-				for obj_id in objs:
-					cls_lookup[obj_id].add(cls)
-		classes = collect
-		
-		if (len(classes) == 1) and (classes[0] not in within_cls_rels):
-			# only one class and no within-class relations apply
-			return set([(obj_id,) for obj_id in objects0])
-		
-		connecting = set()
-		cls0_ = _get_class_members(cls0, leave_names = True)
-		if len(classes) > 1:
-			# collect objects from unspecified classes at shortest paths between specified classes
-			for cls1 in classes[1:]:
-				if (cls0, cls1) in rels:
-					continue
-				cls1_ = _get_class_members(cls1, leave_names = True)
-				connecting.update(self.G.shortest_path_between_classes(cls0_, cls1_))
-			connecting = connecting.difference(classes)
-			if connecting:
-				connecting = set.union(*[_get_class_members(cls) for cls in connecting])
-		
-		paths = set()
-		cmax = len(objects0)
-		cnt = 1
-		if progress is not None:
-			progress.update_state(value = 0, maximum = cmax)
-		for obj_id0 in objects0:
-			if (progress is not None) and (cnt % 100 == 0):
-				if progress.cancel_pressed():
-					return []
-				progress.update_state(value = cnt)
-			cnt += 1
-			paths_ = _get_paths(
-				obj_id0, cls_lookup, rels, within_cls_rels, 
-				asterisk_rels, classes, connecting, mandatory_classes,
-			)
-			if paths_:
-				paths.update(paths_)
-		
-		if progress is not None:
-			progress.update_state(value = cmax)
-		
-		return paths
+		return get_linked_objects(self, classes, relations, progress)
 	
 	def get_query(self, querystr: str, **kwargs) -> object:
 		
@@ -1037,7 +1103,7 @@ class Store(object):
 		# and objects related to them, except objects from the same class as the
 		# specified objects
 		
-		def get_or_add_object(store, obj, obj_lookup):
+		def _get_or_add_object(store, obj, obj_lookup):
 			
 			if obj.id in obj_lookup:
 				return obj_lookup[obj.id]
@@ -1045,13 +1111,18 @@ class Store(object):
 			obj_lookup[obj.id] = obj2
 			return obj2
 		
+		def _get_object_data(obj):
+			data = {descr.name: obj.get_descriptor(descr) for descr in obj.get_descriptors()}
+			locations = {descr.name: obj.get_location(descr) for descr in obj.get_descriptors() if obj.get_location(descr) is not None}
+			return data, locations
+		
 		store = Store(keep_temp = True)
 		
 		# add objects
-		obj_lookup = {}  # {orig_id: DObject in store, ...}
+		obj_lookup = {}	 # {orig_id: DObject in store, ...}
 		classes0 = set()
 		for obj in objects:
-			obj_id = get_or_add_object(store, obj, obj_lookup)
+			obj_id = _get_or_add_object(store, obj, obj_lookup)
 			for cls in obj.get_classes():
 				classes0.add(cls.name)
 		
@@ -1062,30 +1133,29 @@ class Store(object):
 		while queue:
 			orig_id = queue.pop()
 			orig = self.get_object(orig_id)
+			classes1 = set(orig.get_class_names())
 			for orig2, label in orig.get_relations():
-				classes = set([cls.name for cls in orig2.get_classes()])
-				if classes.intersection(classes0):
+				classes2 = set(orig2.get_class_names())
+				if classes2.intersection(classes0 | classes1):
 					continue
-				classes_all.update(classes)
+				classes_all.update(classes2)
 				weight = orig.get_relation_weight(orig2, label)
-				obj2 = get_or_add_object(store, orig2, obj_lookup)
-				obj_lookup[orig_id].add_relation(obj2, label, weight = weight)
+				obj2 = _get_or_add_object(store, orig2, obj_lookup)
 				if orig2.id in done:
 					continue
 				queue.add(orig2.id)
 				done.add(orig2.id)
 		
-		# add within-class relations
-		for orig_id in obj_lookup:
-			orig = self.get_object(orig_id)
-			classes = set([cls.name for cls in orig.get_classes()])
-			for orig2, label in orig.get_relations():
+		# add relations between objects
+		for orig_id_1 in obj_lookup:
+			orig1 = self.get_object(orig_id_1)
+			for orig2, label in orig1.get_relations():
+				if label.startswith('~'):
+					continue
 				if orig2.id not in obj_lookup:
 					continue
-				if not classes.intersection([cls.name for cls in orig2.get_classes()]):
-					continue
-				weight = orig.get_relation_weight(orig2, label)
-				obj_lookup[orig_id].add_relation(obj_lookup[orig2.id], label, weight = weight)
+				weight = orig1.get_relation_weight(orig2.id, label)
+				obj_lookup[orig_id_1].add_relation(obj_lookup[orig2.id], label, weight)
 		
 		# collect descriptor names
 		for orig_id in obj_lookup:
@@ -1122,14 +1192,9 @@ class Store(object):
 		# add descriptors & locations
 		for orig_id in obj_lookup:
 			orig = self.get_object(orig_id)
-			obj = obj_lookup[orig_id]
-			for descr in orig.get_descriptors():
-				value = orig.get_descriptor(descr)
-				self.add_external_descriptor(obj, descr.name, value)
-				location = orig.get_location(descr)
-				if location is not None:
-					obj.set_location(descr.name, location)
-		
+			data, locations = _get_object_data(orig)
+			self.set_object_descriptors(obj_lookup[orig_id], data, locations=locations, match_empty=True, store=self)
+			
 		return store
 	
 	
