@@ -8,6 +8,7 @@ from deposit.datasource.db_rel import DBRel
 from deposit import __version__
 from deposit.utils.fnc_files import (url_to_path)
 from deposit.utils.fnc_serialize import (value_to_str)
+from deposit.store.dresource import DResource
 
 import deposit
 
@@ -204,7 +205,9 @@ def test_resource_descriptor(store):
 		("image1.jpg", "image1.jpg", True, True),
 		("image2.jpg", "image2.jpg", True, True),
 		("noext", "noext", True, False),
-		("test_pdf.pdf", "test_pdf.pdf", True, False),
+		("test_pdf_1.pdf", "test_pdf_1.pdf", True, False),
+		("test_pdf_2.pdf", "test_pdf_2.pdf", True, False),
+		("scsctt.sls", "ščščťť.šľš", True, False),
 	]
 	for name, exp in zip(sorted(pytest.res_names), expected):
 		exp_stored, exp_filename, exp_is_stored, exp_is_image = exp
@@ -329,10 +332,10 @@ def test_geometry_descriptor(store):
 		['SRID=1234;VERT_SRID=5678;POINT(1 2)', 'Point', [1, 2], 1234, 5678],
 		['SRID=1234;MULTIPOINTZ(1 2 7, 3 4 8, 5 6 9)', 'MultiPointZ', [[1, 2, 7], [3, 4, 8], [5, 6, 9]], 1234, None],
 		['LINESTRINGM(1 2 7 10, 3 4 8 11, 5 6 9 12)', 'LineString', [[1, 2, 7, 10], [3, 4, 8, 11], [5, 6, 9, 12]], None, None],
-		['POLYGON((1 2, 3 4, 5 6, 1 2), (7 8, 9 10, 11 12, 7 8), (13 14, 15 16, 17 18, 13 14))', 'POLYGON', [[[1, 2], [3, 4], [5, 6], [1, 2]], [[7, 8], [9, 10], [11, 12], [7, 8]], [[13, 14], [15, 16], [17, 18], [13, 14]]], None, None],
-		['POLYGONZ((1 2 7, 3 4 8, 5 6 9, 1 2 7))', 'PolygonZ', [[[1, 2, 7], [3, 4, 8], [5, 6, 9], [1, 2, 7]]], None, None],
-		['MULTIPOLYGONZ(((1 2 25, 3 4 26, 5 6 27, 1 2 25), (7 8 28, 9 10 29, 11 12 30, 7 8 28)), ((13 14 31, 15 16 32, 17 18 33, 13 14 31), (19 20 34, 21 22 35, 23 24 36, 19 20 34)))', 'MultiPolygonZ', [[[[1, 2, 25], [3, 4, 26], [5, 6, 27], [1, 2, 25]], [[7, 8, 28], [9, 10, 29], [11, 12, 30], [7, 8, 28]]], [[[13, 14, 31], [15, 16, 32], [17, 18, 33], [13, 14, 31]], [[19, 20, 34], [21, 22, 35], [23, 24, 36], [19, 20, 34]]]], None, None],
-		['MULTIPOLYGONM(((1 2 25 37, 3 4 26 18, 5 6 27 39, 1 2 25 37)), ((13 14 31 40, 15 16 32 41, 17 18 33 42, 13 14 31 40), (19 20 34 43, 21 22 35 44, 23 24 36 45, 19 20 34 43)))', 'MULTIPOLYGON', [[[[1, 2, 25, 37], [3, 4, 26, 18], [5, 6, 27, 39], [1, 2, 25, 37]]], [[[13, 14, 31, 40], [15, 16, 32, 41], [17, 18, 33, 42], [13, 14, 31, 40]], [[19, 20, 34, 43], [21, 22, 35, 44], [23, 24, 36, 45], [19, 20, 34, 43]]]], None, None],
+		['POLYGON((1 2, 3 4, 5 6), (7 8, 9 10, 11 12), (13 14, 15 16, 17 18))', 'POLYGON', [[[1, 2], [3, 4], [5, 6]], [[7, 8], [9, 10], [11, 12]], [[13, 14], [15, 16], [17, 18]]], None, None],
+		['POLYGONZ((1 2 7, 3 4 8, 5 6 9))', 'PolygonZ', [[[1, 2, 7], [3, 4, 8], [5, 6, 9]]], None, None],
+		['MULTIPOLYGONZ(((1 2 25, 3 4 26, 5 6 27), (7 8 28, 9 10 29, 11 12 30)), ((13 14 31, 15 16 32, 17 18 33), (19 20 34, 21 22 35, 23 24 36)))', 'MultiPolygonZ', [[[[1, 2, 25], [3, 4, 26], [5, 6, 27]], [[7, 8, 28], [9, 10, 29], [11, 12, 30]]], [[[13, 14, 31], [15, 16, 32], [17, 18, 33]], [[19, 20, 34], [21, 22, 35], [23, 24, 36]]]], None, None],
+		['MULTIPOLYGONM(((1 2 25 37, 3 4 26 18, 5 6 27 39)), ((13 14 31 40, 15 16 32 41, 17 18 33 42), (19 20 34 43, 21 22 35 44, 23 24 36 45)))', 'MULTIPOLYGON', [[[[1, 2, 25, 37], [3, 4, 26, 18], [5, 6, 27, 39]]], [[[13, 14, 31, 40], [15, 16, 32, 41], [17, 18, 33, 42]], [[19, 20, 34, 43], [21, 22, 35, 44], [23, 24, 36, 45]]]], None, None],
 	]
 	for name, exp in zip(pytest.obj4.get_descriptor_names(ordered = True), expected):
 		geo = pytest.obj4.get_descriptor(name)
@@ -396,7 +399,7 @@ def test_del_resource(store):
 	
 	local_folder = store.get_folder()
 	assert len(os.listdir(os.path.join(local_folder, "0000"))) == 0
-	assert len(os.listdir(os.path.join(local_folder, "_deleted"))) == 7
+	assert len(os.listdir(os.path.join(local_folder, "_deleted"))) == 8
 
 def test_add_data_row():
 	
@@ -445,12 +448,100 @@ def test_add_data_row():
 	
 	assert query.columns == [('Feature', 'Name'), ('Area', 'Name')]
 	assert [row for row in query] == [[(2, 'A1.F1'), (1, 'A1')], [(3, 'A1.F2'), (1, 'A1')], [(4, 'A1.F3'), (1, 'A1')], [(5, 'A1.F3'), (1, 'A1')], [(6, 'A1.F4'), (None, None)]]
-	
 
-def test_datasource(store):
+def test_datasource():
+	
+	def setup_store():
+
+		store = deposit.Store()
+		
+		# Setting up classes and their members as described in the provided code
+		finds = store.add_class("Find")
+		features = store.add_class("Feature")
+		areas = store.add_class("Area")
+		sites = store.add_class("Site")
+		
+		sites.add_relation(areas, "contains")
+		areas.add_relation(features, "contains")
+		features.add_relation(finds, "contains")
+		
+		s1 = sites.add_member()
+		s1.set_descriptor("Name", "S1")
+		
+		a1 = areas.add_member()
+		a1.set_descriptor("Name", "A1")
+		a2 = areas.add_member()
+		a2.set_descriptor("Name", "A2")
+		
+		fe11 = features.add_member()
+		fe11.set_descriptor("Name", "A1.F1")
+		fe12 = features.add_member()
+		fe12.set_descriptor("Name", "A1.F2")
+		fe13 = features.add_member()
+		fe13.set_descriptor("Name", "A1.F3")
+		fe14 = features.add_member()
+		fe14.set_descriptor("Name", "A1.F4")
+		fe24 = features.add_member()
+		fe24.set_descriptor("Name", "A2.F4")
+		fe25 = features.add_member()
+		fe25.set_descriptor("Name", "A2.F5")
+		
+		f111 = finds.add_member()
+		f111.set_descriptor("Name", "A1.F1.1")
+		f112 = finds.add_member()
+		f112.set_descriptor("Name", "A1.F1.2")
+		f113 = finds.add_member()
+		f113.set_descriptor("Name", "A1.F1.3")
+		f121 = finds.add_member()
+		f121.set_descriptor("Name", "A1.F2.1")
+		f122 = finds.add_member()
+		f122.set_descriptor("Name", "A1.F2.2")
+		f131 = finds.add_member()
+		f131.set_descriptor("Name", "A1.F3.1")
+		f241 = finds.add_member()
+		f241.set_descriptor("Name", "A2.F4.1")
+		
+		s1.add_relation(a1, "contains")
+		s1.add_relation(a2, "contains")
+		
+		a1.add_relation(fe11, "contains")
+		a1.add_relation(fe12, "contains")
+		a1.add_relation(fe13, "contains")
+		a1.add_relation(fe14, "contains")
+		a2.add_relation(fe24, "contains")
+		a2.add_relation(fe25, "contains")
+		
+		fe11.add_relation(f111, "contains")
+		fe11.add_relation(f112, "contains")
+		fe11.add_relation(f113, "contains")
+		fe12.add_relation(f121, "contains")
+		fe12.add_relation(f122, "contains")
+		fe13.add_relation(f131, "contains")
+		fe24.add_relation(f241, "contains")
+		
+		fe11.add_relation(fe12, "disturbs")
+		fe12.add_relation(fe14, "disturbs")
+		fe11.add_relation(fe13, "covers")
+		fe13.add_relation(fe14, "covers")
+		fe14.add_relation(fe11, "covers")
+		fe24.add_relation(fe25, "disturbs")
+		
+		return store, fe11, fe12, fe24
+	
+	def clean_up(folder):
+		
+		if os.path.isdir(folder):
+			shutil.rmtree(folder)
 	
 	def get_store_data(store):
 		
+		def _get_value(value):
+			if isinstance(value, DResource):
+				if value.url is None:
+					return value_to_str(value)
+				return os.path.split(url_to_path(value.url))[-1]
+			return value_to_str(value)
+		
 		classes = []
 		cls_relations = []
 		objects = []
@@ -473,8 +564,8 @@ def test_datasource(store):
 		for obj in store.get_objects():
 			objects.append([
 				obj.id,
-				natsorted([[descr.name, value_to_str(obj.get_descriptor(descr))] for descr in obj.get_descriptors()]),
-				natsorted([[descr.name, value_to_str(obj.get_location(descr))] for descr in obj.get_descriptors()]),
+				natsorted([[descr.name, _get_value(obj.get_descriptor(descr))] for descr in obj.get_descriptors()]),
+				natsorted([[descr.name, _get_value(obj.get_location(descr))] for descr in obj.get_descriptors()]),
 			])
 			for obj2, label in obj.get_relations():
 				obj_relations.append([obj.id, label, obj2.id])
@@ -505,15 +596,24 @@ def test_datasource(store):
 		cursor.execute("SELECT tablename FROM pg_tables WHERE schemaname = '%s';" % (schema))
 		tables = [row[0] for row in cursor.fetchall()]
 		for table in tables:
-			cursor.execute("DROP TABLE %s.\"%s\";" % (schema, table))
+			cursor.execute("DROP TABLE %s.\"%s\" CASCADE;" % (schema, table))
 		conn.commit()
 		cursor.close()
 		conn.close()
 	
-	pytest.obj1.set_resource_descriptor("Image", "tests//samples dir with spaces//image1.jpg")
-	pytest.obj2.set_resource_descriptor("Image", "tests//samples dir with spaces//image2.jpg")
-	pytest.obj2.set_resource_descriptor("File", "tests//samples dir with spaces//ščščťť.šľš")
-	pytest.obj3.set_resource_descriptor("File", "tests//samples dir with spaces//!@#$%^&() .xx.yy.zz")
+	src_folder = "tests//test_db_src"
+	tgt_folder = "tests//test_db_tgt"
+	
+	clean_up(src_folder)
+	clean_up(tgt_folder)
+	
+	store_src, fe11, fe12, fe24 = setup_store()
+	store_src.set_local_folder(src_folder)
+	
+	fe11.set_resource_descriptor("Image", "tests//samples dir with spaces//image1.jpg")
+	fe12.set_resource_descriptor("Image", "tests//samples dir with spaces//image2.jpg")
+	fe12.set_resource_descriptor("File", "tests//samples dir with spaces//ščščťť.šľš")
+	fe24.set_resource_descriptor("File", "tests//samples dir with spaces//!@#$%^&() .xx.yy.zz")
 	
 	coords_point = [1,2]
 	coords_polygon = [
@@ -523,24 +623,26 @@ def test_datasource(store):
 			[5,6],
 		],
 	]
-	pytest.obj1.set_location("Image", ("POINT", coords_point))
-	pytest.obj2.set_location("Image", ("POLYGON", coords_polygon, 123))
+	fe11.set_location("Image", ("POINT", coords_point))
+	fe12.set_location("Image", ("POLYGON", coords_polygon, 123))
 	
-	data0 = get_store_data(store)
-	prepare_folder(pytest.local_folder)
+	data0 = get_store_data(store_src)
 	
-	store.save(path = os.path.join(pytest.local_folder, "data.json"))
-	store.clear()
-	store.load(path = os.path.join(pytest.local_folder, "data.json"))
+	prepare_folder(tgt_folder)
+	store_src.save(path = os.path.join(tgt_folder, "data.json"))
+	store = deposit.Store()
+	store.load(path = os.path.join(tgt_folder, "data.json"))
 	data = get_store_data(store)
 	assert data == data0
 	
-	store.save(path = os.path.join(pytest.local_folder, "data.pickle"))
-	store.clear()
-	store.load(path = os.path.join(pytest.local_folder, "data.pickle"))
+	prepare_folder(tgt_folder)
+	store_src.save(path = os.path.join(tgt_folder, "data.pickle"))
+	store = deposit.Store()
+	store.load(path = os.path.join(tgt_folder, "data.pickle"))
 	data = get_store_data(store)
 	assert data == data0
 	
+	
 	connstr_db = "postgres://user:1111@127.0.0.1:5432/deposit_test?currentSchema=public"
 	connstr_rel = "postgres://user:1111@127.0.0.1:5432/deposit_test_rel?currentSchema=public"
 	prepare_db(connstr_db)
@@ -550,8 +652,9 @@ def test_datasource(store):
 	datasource.set_connstr(connstr_db)
 	datasource.set_identifier("test1")
 	datasource.create()
-	datasource.save(store)
-	store.clear()
+	datasource.save(store_src)
+	prepare_folder(tgt_folder)
+	store = deposit.Store()
 	store.load(connstr = connstr_db, identifier = "test1")
 	data = get_store_data(store)
 	assert data == data0
@@ -560,12 +663,15 @@ def test_datasource(store):
 	datasource.set_connstr(connstr_rel)
 	datasource.set_identifier("test1")
 	datasource.create()
-	datasource.save(store)
-	store.clear()
+	datasource.save(store_src)
+	prepare_folder(tgt_folder)
+	store = deposit.Store()
 	store.load(connstr = connstr_rel, identifier = "test1")
 	data = get_store_data(store)
 	assert data == data0
 	
 	prepare_db(connstr_db)
 	prepare_db(connstr_rel)
-	prepare_folder(pytest.local_folder)
+	clean_up(src_folder)
+	clean_up(tgt_folder)
+
